import Database from "better-sqlite3";
import * as env from "./env.js";
import path from "path";
import argon2 from "argon2";
import { z } from "zod";

const db = new Database(path.join(env.DataRoot, "control.db"));
db.pragma("journal_mode = WAL");

export async function ensureDatabases() {
    db.exec(`CREATE TABLE IF NOT EXISTS users(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        username NOT NULL UNIQUE,
        passwordHash NOT NULL,
        allowNewUserCreation NOT NULL DEFAULT 0
    )`);

    db.prepare(`INSERT INTO users(username, passwordHash, allowNewUserCreation) VALUES(?, ?, 1)
        ON CONFLICT DO NOTHING`).run("default", await argon2.hash(env.DefaultPassword, {
            timeCost: 5
        }));
        
    db.exec(`CREATE TABLE IF NOT EXISTS sessions(
        id TEXT PRIMARY KEY DEFAULT (lower(hex(randomblob(64)))),
        user INTEGER NOT NULL
    )`);
}


const authResponse = z.object({
    passwordHash: z.string()
});
const sessionResponse = z.object({
    id: z.string()
});
export async function authenticate(username: string, password: string): Promise<string|null> {
    const rawrow = db.prepare("SELECT passwordHash FROM users WHERE username = ?").get(username);
    if (!rawrow) {
        return null;
    }
    
    const row = authResponse.parse(rawrow);
    const valid = await argon2.verify(row.passwordHash, password);
    
    if (!valid) return null;

    const session = sessionResponse.parse(db.prepare("INSERT INTO sessions(user) VALUES(?) RETURNING id").get(username));

    return session.id;
}

const sessionUserResponse = z.object({
    user: z.string()
});
export function sessionValidate(session: string): string|null {
    const rawrow = db.prepare("SELECT user FROM sessions WHERE id = ?").get(session);
    if (!rawrow) return null;

    const row = sessionUserResponse.parse(rawrow);
    return row.user;
}

const canCreateNewUsersResponse = z.object({
    allowNewUserCreation: z.number()
});
export function canCreateNewUsers(username: string): boolean {
    const rawrow = db.prepare("SELECT allowNewUserCreation FROM users WHERE username = ?").get(username);
    if (!rawrow) {
        return false;
    }

    const row = canCreateNewUsersResponse.parse(rawrow);

    return row.allowNewUserCreation === 1;
}

export async function createNewUser(username: string, password: string, allowNewUserCreation: boolean) {
    db.prepare(`INSERT INTO users(username, passwordHash, allowNewUserCreation) VALUES(?, ?, ?)
        ON CONFLICT DO NOTHING`).run(username, await argon2.hash(password, {
            timeCost: 5
        }), allowNewUserCreation ? 1 : 0);
}

export async function updatePassword(username: string, password: string) {
    db.prepare(`UPDATE users SET passwordHash = ? WHERE username = ?`).run(await argon2.hash(password, {
        timeCost: 5
    }), username);
}

export async function deleteSession(sid: string) {
    db.prepare(`DELETE FROM sessions WHERE id = ?`).run(sid);
}