import Database from 'better-sqlite3';
import { readdirSync, readFileSync, mkdirSync } from 'node:fs';
import { join } from 'node:path';
import z from 'zod';
import { DataRoot } from './env.js';

const games_dir = readdirSync(join(DataRoot, 'games'));

const manifest = z.object({
    name: z.string(),
    version: z.string(),
    entrypoint: z.string(),
    localstorage_setup: z.union([
        z.literal("immediate"),
        z.literal("deferred")
    ])
});

const manifestCache: Map<string, z.infer<typeof manifest>> = new Map();

for (let game of games_dir) {
    const manifestContent = manifest.parse(JSON.parse(readFileSync(join(DataRoot, 'games', game, 'manifest.json'), 'utf-8')));
    manifestCache.set(game, manifestContent);
}

export function gameExists(name: string): boolean {
    return manifestCache.has(name);
}

export function getAllGames(): Array<{ slug: string } & z.infer<typeof manifest>> {
    let data = [];
    for (let [k, v] of manifestCache.entries()) {
        data.push({
            slug: k,
            ...v
        })
    }

    return data;
}

export function getGame(name: string): z.infer<typeof manifest> {
    if (!manifestCache.has(name)) throw new Error("no game (ps5 reference)");

    return manifestCache.get(name)!;
}

export function getStorageDatabaseForUser(storageName: string, name: string, user: string): Database.Database {
    if (!manifestCache.has(name)) throw new Error("no game (ps5 reference)");

    mkdirSync(join(DataRoot, "storage", user, name), { recursive: true });

    const db = new Database(join(DataRoot, "storage", user, name, storageName + ".db"));
    db.pragma("journal_mode = WAL");

    db.exec("CREATE TABLE IF NOT EXISTS kv(key TEXT PRIMARY KEY, value TEXT)");

    return db;
}