import express from "express";
import { sessionValidate } from "./auth.js";
import "./game.js";
import { getGame, getStorageDatabaseForUser } from "./game.js";
import path from "path";
import { DataRoot, RootDomain } from "./env.js";
import mime from "mime";
import fs from "fs";
import { injectStandardHeader } from "./injector.js";
import z from "zod";

const router = express.Router();
const preauthRouter = express.Router();
const postauthRouter = express.Router();

const base = import.meta.dirname;

preauthRouter.get("/cgi-bin/rpc/dollhouse_authenticate", (req, res) => {
    if (!req.query.sid) {
        res.status(400).send("Bad Request");
        return;
    }

    const username = sessionValidate(req.query.sid as string);
    if (username === null) {
        res.status(400).send("Bad Request");
        return;
    }

    res.cookie("rdh_sid", req.query.sid as string, {
        httpOnly: true,
        path: "/",
        maxAge: 86400 * 400 * 1000
    });

    res.set('Location', '/cgi-bin/startup');
    res.status(307).send("Welcome");
});

postauthRouter.get("/cgi-bin/startup", (req, res) => {
    const game = getGame(req.game);
    
    res.set('Location', `/${game.entrypoint}`);
    res.status(307).send("off you go girl");
});

postauthRouter.get("/cgi-bin/rpc/dollhouse_authenticate", (req, res) => {
    const game = getGame(req.game);
    
    res.set('Location', `/${game.entrypoint}`);
    res.status(307).send("off you go girl");
});

postauthRouter.use("/cgi-bin/static", express.static(path.join(base, "injections")));

postauthRouter.get("/cgi-bin/rpc/storage/:name/read_all", (req, res) => {
    if (!(["localStorage", "sessionStorage"]).includes(req.params.name)) {
        res.status(400).send("Bad Request");
        return;
    }
    
    const db = getStorageDatabaseForUser(req.params.name, req.game, req.active_user);

    const results = db.prepare("SELECT * FROM kv").all();
    db.close();
    res.json(results);
});

const upsertRequest = z.object({
    key: z.string(),
    value: z.string()
});

postauthRouter.post("/cgi-bin/rpc/storage/:name/upsert", express.json({
    limit: 1024 * 1024 * 1024 * 16,
    type: () => true
}), (req, res) => {
    if (!(["localStorage", "sessionStorage"]).includes(req.params.name)) {
        res.status(400).send("Bad Request");
        return;
    }
    
    const db = getStorageDatabaseForUser(req.params.name, req.game, req.active_user);
    const body = upsertRequest.parse(req.body);

    console.log("It is quite upserting", req.params.name, body.key);

    db.prepare("INSERT INTO kv(key, value) VALUES(?, ?) ON CONFLICT (key) DO UPDATE SET value = ? WHERE key = ?").run(
        body.key,
        body.value,
        body.value,
        body.key
    );
    db.close();

    res.json({ ok: true, "or is it?": "i don't know" });
});

const removeRequest = z.object({
    key: z.string()
});

postauthRouter.post("/cgi-bin/rpc/storage/:name/remove", express.json({
    type: () => true
}), (req, res) => {
    if (!(["localStorage", "sessionStorage"]).includes(req.params.name)) {
        res.status(400).send("Bad Request");
        return;
    }
    
    const db = getStorageDatabaseForUser(req.params.name, req.game, req.active_user);
    const body = removeRequest.parse(req.body);

    db.prepare("DELETE FROM kv WHERE key = ?").run(
        body.key
    );
    db.close();

    res.json({ ok: true });
});

postauthRouter.post("/cgi-bin/rpc/storage/:name/clear", (req, res) => {
    if (!(["localStorage", "sessionStorage"]).includes(req.params.name)) {
        res.status(400).send("Bad Request");
        return;
    }

    const db = getStorageDatabaseForUser(req.params.name, req.game, req.active_user);

    db.prepare("DELETE FROM kv").run();
    db.close();

    res.json({ ok: true, "or is it?": "is it ever ok" });
});

postauthRouter.use((req, res) => {
    const game = getGame(req.game);

    const parsedUrl = URL.parse(req.url, `https://${RootDomain}`);
    if (!parsedUrl) {
        res.status(400).send("Bad Request");
        return;
    }
    if (parsedUrl.pathname === "/") {
        res.set('Location', '/cgi-bin/startup');
        res.status(307).send("Welcome");
        return;
    }
    const pathname = decodeURI(parsedUrl.pathname);
    const resolved = path.resolve("/", pathname);
    const realFile = path.join(DataRoot,'games', req.game, 'data', resolved);
    console.log(realFile);

    if (!fs.existsSync(realFile)) {
        res.status(404).send("Not Found");
        return;
    }

    res.set("Content-Type", mime.getType(realFile) || "application/octet-stream");
    
    // TODO: Perform injection into the entrypoint

    if (realFile.endsWith(game.entrypoint)) {
        injectStandardHeader(realFile, res);
        return;
    }

    // Fallback path
    const size = fs.statSync(realFile);
    res.set("Content-Length", size.size.toString());
    const stream = fs.createReadStream(realFile);
    stream.pipe(res);
});

router.use((req, res, next) => {
    const session = req.cookies["rdh_sid"];
    if (!session) {
        return preauthRouter(req, res, next);
    }

    const username = sessionValidate(session);
    if (username === null) {
        return preauthRouter(req, res, next);
    }
    
    req.active_user = username;
    return postauthRouter(req, res, next);
});

export default router;