import express, { urlencoded } from "express";
import { join } from "path";
import z from "zod";
import { authenticate, canCreateNewUsers, createNewUser, deleteSession, sessionValidate, updatePassword } from "./auth.js";
import { getAllGames } from "./game.js";
import { Protocol, RootDomain } from "./env.js";

const router = express.Router();
const preauthRouter = express.Router();
const postauthRouter = express.Router();

const base = import.meta.dirname;

const loginPayload = z.object({
    username: z.string(),
    password: z.string()
});

preauthRouter.use(urlencoded());
preauthRouter.get("/", (req, res) => {
    res.sendFile(join(base, "statics", "login.html"));
});

preauthRouter.post("/", async (req, res) => {
    const body = loginPayload.parse(req.body);
    const result = await authenticate(body.username, body.password);
    
    if (result === null) {
        res.status(403).send("Bad username or password");
        return;
    }

    res.cookie("rdh_sid", result, {
        httpOnly: true,
        path: "/",
        maxAge: 86400 * 400 * 1000
    });

    res.set('Location', '/');
    res.status(303).send("Welcome");
});

postauthRouter.get("/", (req, res) => {
    let html = `<h1>Available games</h1><ul>`;
    for (let game of getAllGames()) {
        html = html + `<li><a href="${Protocol}://${game.slug}.${RootDomain}/cgi-bin/rpc/dollhouse_authenticate?sid=${req.cookies["rdh_sid"]}">${game.name}</a></li>`
    }
    html = html + "</ul>"

    if (canCreateNewUsers(req.active_user)) {
        html = html + `<br><h1>Create a new user</h1><form action='/create_user' method='POST'>
            <label>Username: <input type="username" name="username"></label><br/>
            <label>Password: <input type="password" name="password"></label><br/>
            <label>Allow new user creation: <input type="checkbox" name="allow"></label><br/>
            <input type="submit">
        </form>`
    }

    html = html + `<br><h1>Change password</h1><form action='/change_password' method='POST'>
        <label>New Password: <input type="password" name="password"></label><br/>
        <input type="submit">
    </form>
    <form action='/delete_session' method='POST'>
        <input type="hidden" name="deletion" value="delete">
        <input type="submit" value="Delete Session">
    </form>`
    res.set("Content-Type", "text/html;charset=UTF-8");
    res.send(html);
});

const createPayload = z.object({
    username: z.string(),
    password: z.string(),
    allow: z.optional(z.literal("on"))
});
postauthRouter.post('/create_user', urlencoded(), (req, res) => {
    if (!canCreateNewUsers(req.active_user)) {
        res.status(401).send("Unauthorized");
        return;
    }

    const body = createPayload.parse(req.body);
    const allow = body.allow === "on";

    createNewUser(body.username, body.password, allow);

    res.set('Location', '/');
    res.status(303).send("Welcome");
});

const changePasswordPayload = z.object({
    password: z.string()
});
postauthRouter.post('/change_password', urlencoded(), (req, res) => {
    const body = changePasswordPayload.parse(req.body);

    updatePassword(req.active_user, body.password);

    res.set('Location', '/');
    res.status(303).send("Welcome");
})

const deleteSessionPayload = z.object({
    deletion: z.literal('delete')
});
postauthRouter.post('/delete_session', urlencoded(), (req, res) => {
    deleteSessionPayload.parse(req.body);

    deleteSession(req.cookies["rdh_sid"]);

    res.set('Location', '/');
    res.status(303).send("Welcome");
})


router.use((req, res, next) => {
    const session = req.cookies["rdh_sid"];
    if (!session) {
        return preauthRouter(req, res, next);
    }

    const username = sessionValidate(session);
    if (username === null) {
        return preauthRouter(req, res, next);
    }
    
    req.active_user = username;
    return postauthRouter(req, res, next);
});

export default router;